@@ -234,6 +234,20 @@ namespace gbe {
     }
 
     DenseMap<Value*, VectorValues> vectorVals;
+    struct VecValElement{
+      VecValElement(VectorValues *v, uint32_t i) : vecVals(v), id(i) {}
+      VectorValues *vecVals;
+      uint32_t id;
+    };
+    DenseMap<Value*, SmallVector<VecValElement, 16>> usedVecVals;
+
+    void setComponent(VectorValues &vecVals, uint32_t c, llvm::Value* val) {
+      vecVals.setComponent(c, val);
+      usedVecVals[val].push_back(VecValElement(&vecVals, c));
+    }
+
+    void replaceAllUsesOfWith(Instruction* from, Instruction* to);
+
     Module* module;
     IRBuilder<>* builder;
 
@@ -277,8 +291,17 @@ namespace gbe {
 
   bool Scalarize::IsPerComponentOp(const Instruction* inst)
   {
-    //if (const IntrinsicInst* intr = dyn_cast<const IntrinsicInst>(inst))
-    //    return IsPerComponentOp(intr);
+    if (const IntrinsicInst* intr = dyn_cast<const IntrinsicInst>(inst))
+    {
+        const Intrinsic::ID intrinsicID = (Intrinsic::ID) intr->getIntrinsicID();
+        switch (intrinsicID) {
+          default: return false;
+          case Intrinsic::sqrt:
+          case Intrinsic::ceil:
+          case Intrinsic::trunc:
+              return true;
+        }
+    }
 
     if (inst->isTerminator())
         return false;
@@ -423,13 +446,17 @@ namespace gbe {
       // assumption. This is due to how getDeclaration operates; it only takes
       // a list of types that fit overloadable slots.
       SmallVector<Type*, 8> tys(1, GetBasicType(inst->getType()));
+
       // Call instructions have the decl as a last argument, so skip it
+      SmallVector<Value*, 8> _args;
+
       for (ArrayRef<Value*>::iterator i = args.begin(), e = args.end() - 1; i != e; ++i) {
         tys.push_back(GetBasicType((*i)->getType()));
+        _args.push_back(*i);
       }
 
       Function* f = Intrinsic::getDeclaration(module, intr->getIntrinsicID(), tys);
-      return CallInst::Create(f, args);
+      return CallInst::Create(f, _args);
     }
 
     NOT_IMPLEMENTED; //gla::UnsupportedFunctionality("Currently unsupported instruction: ", inst->getOpcode(),
@@ -448,7 +475,7 @@ namespace gbe {
       callArgs.push_back(ConstantInt::get(intTy, i));
 
       res = builder->CreateCall(f, callArgs);
-      vVals.setComponent(i, res);
+      setComponent(vVals, i, res);
     }
   }
 
@@ -468,7 +495,7 @@ namespace gbe {
 
       Instruction* res = createScalarInstruction(inst, callArgs);
 
-      vVals.setComponent(i, res);
+      setComponent(vVals, i, res);
       builder->Insert(res);
     }
   }
@@ -549,7 +576,7 @@ namespace gbe {
       int select = sv->getMaskValue(i);
 
       if (select < 0) {
-        vVals.setComponent(i, UndefValue::get(GetBasicType(sv->getOperand(0))));
+        setComponent(vVals, i, UndefValue::get(GetBasicType(sv->getOperand(0))));
         continue;
       }
 
@@ -564,7 +591,7 @@ namespace gbe {
         selectee = sv->getOperand(1);
       }
 
-      vVals.setComponent(i, getComponent(select, selectee));
+      setComponent(vVals, i, getComponent(select, selectee));
     }
 
     return true;
@@ -611,7 +638,7 @@ namespace gbe {
     for (int i = 0; i < GetComponentCount(insn); ++i) {
       Value *cv = ConstantInt::get(intTy, i);
       Value *EI = builder->CreateExtractElement(insn, cv);
-      vVals.setComponent(i, EI);
+      setComponent(vVals, i, EI);
     }
   }
 
@@ -632,7 +659,18 @@ namespace gbe {
   bool Scalarize::scalarizeFuncCall(CallInst* call) {
     if (Function *F = call->getCalledFunction()) {
       if (F->getIntrinsicID() != 0) {   //Intrinsic functions
-        NOT_IMPLEMENTED;
+        const Intrinsic::ID intrinsicID = (Intrinsic::ID) F->getIntrinsicID();
+
+        switch (intrinsicID) {
+          default: GBE_ASSERTM(false, "Unsupported Intrinsic");
+          case Intrinsic::sqrt:
+          case Intrinsic::ceil:
+          case Intrinsic::trunc:
+          {
+            scalarizePerComponent(call);
+          }
+          break;
+        }
       } else {
         Value *Callee = call->getCalledValue();
         const std::string fnName = Callee->getName();
@@ -643,6 +681,7 @@ namespace gbe {
         CallSite::arg_iterator CI = CS.arg_begin() + 1;
 
         switch (genIntrinsicID) {
+          case GEN_OCL_NOT_FOUND:
           default: break;
           case GEN_OCL_READ_IMAGE_I:
           case GEN_OCL_READ_IMAGE_UI:
@@ -679,6 +718,7 @@ namespace gbe {
       setAppendPoint(bt);
       extractFromVector(bt);
     }
+
     return false;
   }
 
@@ -694,6 +734,16 @@ namespace gbe {
     return false;
   }
 
+  void Scalarize::replaceAllUsesOfWith(Instruction* from, Instruction* to) {
+    GBE_ASSERT(from != NULL);
+    if (from == to)
+      return;
+    for (auto &it : usedVecVals[from])
+      setComponent(*(it.vecVals), it.id, to);
+    usedVecVals[from].clear();
+    from->replaceAllUsesWith(to);
+  }
+
   bool Scalarize::scalarizeExtract(ExtractElementInst* extr)
   {
     //     %res = extractelement <n X ty> %foo, %i
@@ -710,7 +760,7 @@ namespace gbe {
     Value* v = getComponent(component, extr->getOperand(0));
     if(extr == v)
       return false;
-    extr->replaceAllUsesWith(v);
+    replaceAllUsesOfWith(dyn_cast<Instruction>(extr), dyn_cast<Instruction>(v));
 
     return true;
   }
@@ -730,8 +780,8 @@ namespace gbe {
 
     VectorValues& vVals = vectorVals[ins];
     for (int i = 0; i < GetComponentCount(ins); ++i) {
-      vVals.setComponent(i, i == component ? ins->getOperand(1)
-                                           : getComponent(i, ins->getOperand(0)));
+      setComponent(vVals, i, i == component ? ins->getOperand(1)
+                   : getComponent(i, ins->getOperand(0)));
     }
 
     return true;
@@ -765,6 +815,7 @@ namespace gbe {
 #else
     case CallingConv::C:
     case CallingConv::Fast:
+    case CallingConv::SPIR_KERNEL:
 #endif
       break;
     default:
@@ -815,9 +866,9 @@ namespace gbe {
     }
 
     dce();
-
     incompletePhis.clear();
     vectorVals.clear();
+    usedVecVals.clear();
 
     delete builder;
     builder = 0;
