@@ -46,16 +46,18 @@ cl_event_is_gpu_command_type(cl_command_type type)
   }
 }
 
-void cl_event_flush(cl_event event)
+int cl_event_flush(cl_event event)
 {
+  int err = CL_SUCCESS;
   assert(event->gpgpu_event != NULL);
   if (event->gpgpu) {
-    cl_command_queue_flush_gpgpu(event->queue, event->gpgpu);
+    err = cl_command_queue_flush_gpgpu(event->queue, event->gpgpu);
     cl_gpgpu_delete(event->gpgpu);
     event->gpgpu = NULL;
   }
   cl_gpgpu_event_flush(event->gpgpu_event);
-  event->queue->last_event = event;
+  set_last_event(event->queue, event);
+  return err;
 }
 
 cl_event cl_event_new(cl_context ctx, cl_command_queue queue, cl_command_type type, cl_bool emplict)
@@ -115,20 +117,11 @@ void cl_event_delete(cl_event event)
   if (atomic_dec(&event->ref_n) > 1)
     return;
 
-  if(event->queue && event->queue->last_event == event)
-    event->queue->last_event = NULL;
+  if(event->queue && get_last_event(event->queue) == event)
+    set_last_event(event->queue, NULL);
 
   /* Call all user's callback if haven't execute */
-  user_callback *cb = event->user_cb;
-  while(event->user_cb) {
-    cb = event->user_cb;
-    if(cb->executed == CL_FALSE) {
-      cb->executed = CL_TRUE;
-      cb->pfn_notify(event, event->status, cb->user_data);
-    }
-    event->user_cb = cb->next;
-    cl_free(cb);
-  }
+  cl_event_call_callback(event, CL_COMPLETE, CL_TRUE); // CL_COMPLETE status will force all callbacks that are not executed to run
 
   /* delete gpgpu event object */
   if(event->gpgpu_event)
@@ -180,8 +173,22 @@ cl_int cl_event_set_callback(cl_event event ,
   cb->status      = command_exec_callback_type;
   cb->executed    = CL_FALSE;
 
-  cb->next        = event->user_cb;
-  event->user_cb  = cb;
+
+  // It is possible that the event enqueued is already completed.
+  // clEnqueueReadBuffer can be synchronous and when the callback
+  // is registered after, it still needs to get executed.
+  pthread_mutex_lock(&event->ctx->event_lock); // Thread safety required: operations on the event->status can be made from many different threads
+  if(event->status <= command_exec_callback_type) {
+    /* Call user callback */
+    pthread_mutex_unlock(&event->ctx->event_lock); // pfn_notify can call clFunctions that use the event_lock and from here it's not required
+    cb->pfn_notify(event, event->status, cb->user_data);
+    cl_free(cb);
+  } else {
+    // Enqueue to callback list
+    cb->next        = event->user_cb;
+    event->user_cb  = cb;
+    pthread_mutex_unlock(&event->ctx->event_lock);
+  }
 
 exit:
   return err;
@@ -214,7 +221,7 @@ cl_int cl_event_check_waitlist(cl_uint num_events_in_wait_list,
       err = CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST;
       goto exit;
     }
-    if(event && *event == event_wait_list[i])
+    if(event && event == &event_wait_list[i])
       goto error;
     if(event_wait_list[i]->ctx != ctx)
       goto error;
@@ -388,9 +395,48 @@ error:
   goto exit;
 }
 
+void cl_event_call_callback(cl_event event, cl_int status, cl_bool free_cb) {
+  user_callback *user_cb = NULL;
+  user_callback *queue_cb = NULL; // For thread safety, we create a queue that holds user_callback's pfn_notify contents
+  user_callback *temp_cb = NULL;
+  user_cb = event->user_cb;
+  pthread_mutex_lock(&event->ctx->event_lock);
+  while(user_cb) {
+    if(user_cb->status >= status
+        && user_cb->executed == CL_FALSE) { // Added check to not execute a callback when it was already handled
+      user_cb->executed = CL_TRUE;
+      temp_cb = cl_malloc(sizeof(user_callback));
+      if(!temp_cb) {
+        break; // Out of memory
+      }
+      temp_cb->pfn_notify = user_cb->pfn_notify; // Minor struct copy to call ppfn_notify out of the pthread_mutex
+      temp_cb->user_data = user_cb->user_data;
+      if(free_cb) {
+        cl_free(user_cb);
+      }
+      if(!queue_cb) {
+        queue_cb = temp_cb;
+        queue_cb->next = NULL;
+      } else { // Enqueue First
+        temp_cb->next = queue_cb;
+        queue_cb = temp_cb;
+      }
+    }
+    user_cb = user_cb->next;
+  }
+  pthread_mutex_unlock(&event->ctx->event_lock);
+
+  // Calling the callbacks outside of the event_lock is required because the callback can call cl_api functions and get deadlocked
+  while(queue_cb) { // For each callback queued, actually execute the callback
+    queue_cb->pfn_notify(event, event->status, queue_cb->user_data);
+    temp_cb = queue_cb;
+    queue_cb = queue_cb->next;
+    cl_free(temp_cb);
+  }
+}
+
 void cl_event_set_status(cl_event event, cl_int status)
 {
-  user_callback *user_cb;
   cl_int ret, i;
   cl_event evt;
 
@@ -437,14 +483,7 @@ void cl_event_set_status(cl_event event, cl_int status)
   pthread_mutex_unlock(&event->ctx->event_lock);
 
   /* Call user callback */
-  user_cb = event->user_cb;
-  while(user_cb) {
-    if(user_cb->status >= status) {
-      user_cb->executed = CL_TRUE;
-      user_cb->pfn_notify(event, event->status, user_cb->user_data);
-    }
-    user_cb = user_cb->next;
-  }
+  cl_event_call_callback(event, status, CL_FALSE);
 
   if(event->type == CL_COMMAND_USER) {
     /* Check all defer enqueue */
@@ -529,9 +568,9 @@ cl_int cl_event_marker_with_wait_list(cl_command_queue queue,
     return CL_SUCCESS;
   }
 
-  if(queue->last_event && queue->last_event->gpgpu_event) {
-    cl_gpgpu_event_update_status(queue->last_event->gpgpu_event, 1);
-  }
+  cl_event last_event = get_last_event(queue);
+  if(last_event && last_event->gpgpu_event)
+    cl_gpgpu_event_update_status(last_event->gpgpu_event, 1);
 
   cl_event_set_status(e, CL_COMPLETE);
   return CL_SUCCESS;
@@ -566,14 +605,69 @@ cl_int cl_event_barrier_with_wait_list(cl_command_queue queue,
     return CL_SUCCESS;
   }
 
-  if(queue->last_event && queue->last_event->gpgpu_event) {
-    cl_gpgpu_event_update_status(queue->last_event->gpgpu_event, 1);
-  }
+  cl_event last_event = get_last_event(queue);
+  if(last_event && last_event->gpgpu_event)
+    cl_gpgpu_event_update_status(last_event->gpgpu_event, 1);
 
   cl_event_set_status(e, CL_COMPLETE);
   return CL_SUCCESS;
 }
 
+cl_ulong cl_event_get_cpu_timestamp(cl_ulong *cpu_time)
+{
+  struct timespec ts;
+
+ if(clock_gettime(CLOCK_MONOTONIC_RAW,&ts) != 0){
+  printf("CPU Timmer error\n");
+  return CL_FALSE;
+  }
+  *cpu_time = (1000000000.0) * (cl_ulong) ts.tv_sec + (cl_ulong) ts.tv_nsec;
+
+  return CL_SUCCESS;
+}
+
+cl_int cl_event_get_queued_cpu_timestamp(cl_event event)
+{
+  cl_int ret_val;
+
+  ret_val = cl_event_get_cpu_timestamp(&event->queued_timestamp);
+
+  return ret_val;
+}
+
+cl_ulong cl_event_get_timestamp_delta(cl_ulong start_timestamp,cl_ulong end_timestamp)
+{
+  cl_ulong ret_val;
+
+  if(end_timestamp > start_timestamp){
+   ret_val = end_timestamp - start_timestamp;
+   }
+  else {
+   /*if start time stamp is greater than end timstamp then set ret value to max*/
+   ret_val = ((cl_ulong) 1 << 32);
+  }
+
+  return ret_val;
+}
+
+cl_ulong cl_event_get_start_timestamp(cl_event event)
+{
+  cl_ulong ret_val;
+
+   ret_val = cl_event_get_timestamp_delta(event->timestamp[0],event->timestamp[2]);
+
+  return ret_val;
+}
+
+cl_ulong cl_event_get_end_timestamp(cl_event event)
+{
+ cl_ulong ret_val;
+
+  ret_val = cl_event_get_timestamp_delta(event->timestamp[0],event->timestamp[3]);
+
+  return ret_val;
+}
+
 cl_int cl_event_get_timestamp(cl_event event, cl_profiling_info param_name)
 {
   cl_ulong ret_val = 0;
